/*
 * Jamote - JAva MOdel-based black box TEsting (http://www.assembla.com/wiki/show/jamote)
 * Copyright (C) 2008 Jamote Team (http://www.assembla.com/spaces/jamote/team)
 * 
 *  This program is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * Jamote utilizes several third-party packages whose code may
 * be distributed under a different license, see the file named LICENSE for
 * more details.
 */


package it.dis.uniroma1.jamote.core;

import it.dis.uniroma1.jamote.core.language.Function;
import it.dis.uniroma1.jamote.core.language.InterfaceParser;
import it.dis.uniroma1.jamote.core.language.JamoteLanguageException;
import it.dis.uniroma1.jamote.core.language.LanguageFactory;
import it.dis.uniroma1.jamote.core.language.Tester;
import it.dis.uniroma1.jamote.core.language.java.JavaLanguageFactory;
import it.dis.uniroma1.jamote.core.modeling.InputModel;
import it.dis.uniroma1.jamote.core.modeling.OutputModel;
import it.dis.uniroma1.jamote.core.modeling.Solver;
import it.dis.uniroma1.jamote.core.modeling.ModelingFactory;
import it.dis.uniroma1.jamote.core.modeling.alloy.AlloyModelingFactory;
import it.dis.uniroma1.jamote.utils.ReportHandler;
import it.dis.uniroma1.jamote.utils.Reporter;
import it.dis.uniroma1.jamote.utils.ReportEntry.Level;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;



/**
 * Classe controller, fornisce un punto d'accesso unico alle funzionalità del core
 *
 */
public class Test {
	
	//
	// Variabili di input
	//
	protected String language;		//Linguaggio di programmazione
	protected String modelingTool;	//Strumento di modellazione
	protected Reporter reporter;
	
	//
	// Variabili per utilizzo interno alla classe
	//
	protected LanguageFactory languageFactory;
	protected ModelingFactory modelingFactory;
	
	
	/**
	 * 
	 * @param language
	 * @param modelingTool
	 * @throws JamoteException
	 */
	public Test(String language, String modelingTool, ReportHandler reportHandler)
			throws JamoteException{
		if(language == null)
			throw new JamoteException("Language not supported: " + language);
		
		if(language.equals("java"))
		{
			this.language = language;
			languageFactory = new JavaLanguageFactory();
		}
		else
			throw new JamoteException("Programming language not supported: "+ language);
		
		if(modelingTool.equals("alloy"))
		{
			this.modelingTool = modelingTool;
			modelingFactory = new AlloyModelingFactory();
		}
		else
			throw new JamoteException("Modeling tool not supported: "+ modelingTool);
		
		this.reporter = new Reporter(reportHandler);
	}
	
	/**
	 * 
	 * @return
	 */
	public List<Function> searchFunctions(String filename) throws JamoteLanguageException{
		assert(languageFactory != null);
		InterfaceParser ifp = languageFactory.createInterfaceParser(this.reporter);
		return ifp.parseFile(filename);
	}

	/**
	 * 
	 * @throws JamoteException
	 */
	public Map<String, List<TestResult> > executeTest(String precoditions, 
			List<EquivalencePartition> partitions, String defines, Function function) 
			throws JamoteException {
		
		assert(languageFactory != null && modelingFactory != null);
		Map<String, List<TestResult> > results = new HashMap<String, List<TestResult>>();
		
		Solver solver = modelingFactory.createSolver(this.reporter);
		Tester tester = languageFactory.createTester(this.reporter);
		
		
		//Per ogni classe di equivalenza
		for(EquivalencePartition partition :partitions){
			int ntests = 0,ncorrect = 0, nout = 0;
			List<TestResult> partitionResults = new ArrayList<TestResult>();
			
			//Genero i modelli astratti della classe di equivalenza
			reporter.log(Level.DEFAULT, "---> Testing \""+partition.getName()+
					"\" equivalence partition <---");
			reporter.log(Level.DEFAULT, "Generating models...");
			
			List<InputModel> models = solver.generateModels(function, defines , precoditions, partition);
			
			reporter.log(Level.DEFAULT, "Generation of models completed. " 
					+ models.size() + " models generated");
			
			//Eseguo i test con i modelli trovati
			reporter.log(Level.DEFAULT, "Now testing the function...");
			for(InputModel inputModel: models){
				//ora eseguo il test e prendo il risultato
				reporter.log(Level.DEFAULT, "Testing with input: " + inputModel);
				
				ntests++;
				OutputModel outputModel = tester.executeTest(function, inputModel);
				
				reporter.log(Level.DEFAULT, "Test output: " + outputModel);
				
				//Verifico la correttezza del risultato
				reporter.log(Level.VERBOSE, "Now checking correctness...");
				
				int successful = solver.checkOutput(function, defines,
						partition, inputModel, outputModel);
				
				if(successful == 0)
				{
					ncorrect++;
					reporter.log(Level.DEFAULT, "Test result is correct :)");
				}
				else if(successful == 1)
				{
					reporter.log(Level.WARNING, "Test result is NOT correct :( . " +
							"Input Model: " + inputModel +", Return value: " + outputModel);
					reporter.log(Level.WARNING, "Now will compute the expected output...");
					OutputModel expOut = solver.getExpectedOutput(function, defines, 
							partition, inputModel);
					reporter.log(Level.WARNING, "Expected output is: "+ expOut);
				}
				else if(successful == 2){
					getReporter().log(Level.WARNING,"Return value is outside integer limits! Cannot" +
						" check function correctness!");
					nout++;
				}
				
				
				//Creo il test result
				TestResult tr = new TestResult(inputModel, outputModel,
						successful);
				
				partitionResults.add(tr);
			}
			
			//aggiungo il risultato alla lista dei risultati relativi ad una partizione
			results.put(partition.name, partitionResults);
			
			reporter.log(Level.DEFAULT, "TEST COMPLETED! "+ntests+" were executed. " 
					+ ncorrect + " correct, " + (ntests - ncorrect - nout) + " NOT correct" + 
					", " + nout + " outside integer limits");
		}
		
		return results;
	}


	public String getLanguage() {
		return language;
	}

	public String getModelingTool() {
		return modelingTool;
	}

	public Reporter getReporter() {
		return reporter;
	}
}
