/*
 * Jamote - JAva MOdel-based black box TEsting (http://www.assembla.com/wiki/show/jamote)
 * Copyright (C) 2008 Jamote Team (http://www.assembla.com/spaces/jamote/team)
 * 
 *  This program is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * Jamote utilizes several third-party packages whose code may
 * be distributed under a different license, see the file named LICENSE for
 * more details.
 */


package it.dis.uniroma1.jamote.core.modeling.alloy;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Scanner;

import edu.mit.csail.sdg.alloy4.A4Reporter;
import edu.mit.csail.sdg.alloy4.Err;
import edu.mit.csail.sdg.alloy4.ErrorWarning;
import edu.mit.csail.sdg.alloy4.SafeList;
import edu.mit.csail.sdg.alloy4compiler.ast.Command;
import edu.mit.csail.sdg.alloy4compiler.ast.Sig;
import edu.mit.csail.sdg.alloy4compiler.ast.Type;
import edu.mit.csail.sdg.alloy4compiler.ast.Sig.Field;
import edu.mit.csail.sdg.alloy4compiler.ast.Sig.PrimSig;
import edu.mit.csail.sdg.alloy4compiler.parser.CompUtil;
import edu.mit.csail.sdg.alloy4compiler.parser.Module;
import edu.mit.csail.sdg.alloy4compiler.translator.A4Options;
import edu.mit.csail.sdg.alloy4compiler.translator.A4Solution;
import edu.mit.csail.sdg.alloy4compiler.translator.A4Tuple;
import edu.mit.csail.sdg.alloy4compiler.translator.A4TupleSet;
import edu.mit.csail.sdg.alloy4compiler.translator.TranslateAlloyToKodkod;
import it.dis.uniroma1.jamote.core.EquivalencePartition;
import it.dis.uniroma1.jamote.core.language.Function;
import it.dis.uniroma1.jamote.core.modeling.AbstractObject;
import it.dis.uniroma1.jamote.core.modeling.InputModel;
import it.dis.uniroma1.jamote.core.modeling.OutputModel;
import it.dis.uniroma1.jamote.core.modeling.Solver;
import it.dis.uniroma1.jamote.core.modeling.JamoteModelingException;
import it.dis.uniroma1.jamote.utils.Config;
import it.dis.uniroma1.jamote.utils.Reporter;
import it.dis.uniroma1.jamote.utils.ReportEntry.Level;


/**
 * E' il generatore di modelli astratto basato su Alloy
 *
 */
public class AlloySolver extends Solver{
	public static String InputSig = "Input";
	public static String OutputSig = "Output";
	public static String ExcSig = "Exception";
	public static String MetaSig = "Metadata";
	public static String ArgField = "arg";
	public static String MaxIntField = "maxInt";
	public static String MinIntField = "minInt";
	public static String ModelSizeField = "modelSize";
	public static String RetField = "return";
	public static String ExcField = "exception";
	
	
	public AlloySolver(Reporter reporter)
	{
		super(reporter);
	}
		
	
	/**
	 * @see Solver
	 */
	public List<InputModel> generateModels(Function function, String defines, String preconditions,
			EquivalencePartition partition) throws JamoteModelingException {
		File file = null;
		try{
			long init = System.currentTimeMillis();
			getReporter().log(Level.VERBOSE, "Generating models for partition \"" +
					partition.getName() + "\"");
			
			//Creo il file di input da passare ad Alloy
			file = this.createAlloyInputFile(function, defines, preconditions, partition);
			
			getReporter().log(Level.DEBUG, "Generated Input file:\n--------\n"+
					getFileContents(file,false));
			
			//Configuro Alloy ed eseguo lo script appena creato
			List<InputModel> ret = executeAlloyToFindModels(file,function, partition);
			getReporter().log(Level.VERBOSE, "Model generation took " + (System.currentTimeMillis() - init)
					+ " ms");
			return ret;
		}
		catch(JamoteModelingException je){
			getReporter().log(Level.ERROR, "An error occurred during model generation...aborting", je);
			
			//Stampiamo il contenuto del file
			getReporter().log(Level.ERROR, "Printing Alloy input file: \n" + 
					getFileContents(file,true));
			
			throw je;
		}
		catch(Exception e){
			getReporter().log(Level.ERROR, "Error during model generation...aborting", e);
			throw new JamoteModelingException("Error during model generation", e);
		}
		finally{
			//Elimino il file temporaneo
			if(file != null)
				file.delete();
		}
	}
	

	/**
	 * Crea il file di input necessario a generare i modelli da passare in 
	 * input alla funzione
	 * 
	 * @param function
	 * @param preconditions
	 * @param partition
	 * @return
	 */
	private File createAlloyInputFile(Function function,
			String defines, String preconditions, EquivalencePartition partition) 
			throws JamoteModelingException, IOException {
		
		String contents = "";
		File file = null;
		
		
		getReporter().log(Level.VERBOSE, "Creating Alloy input file...");
		
		//genero l'header
		contents += this.generateHeader(function);
		
		contents += this.generateMetaSig(partition);
		
		//genero le sig di supporto
		contents += this.generateSupportSigs(function);
		
		//Genero le sigs a partire dalla segnatura della funzione da testare
		contents += this.generateInputSig(function);
		
		//aggiungo i defines
		contents += defines + "\n\n";
		
		//Genero i fact corrispondenti alle precondizioni
		contents += this.generatePreconditions(preconditions);
		
		//Genero i pred corrispondenti alle specifiche di input delle classi di equivalenza
		contents += this.generateInputSpecification(partition);
		
		//Genero l'istruzione per la run
		contents += this.generateModelGenerationRun(partition);
		
		//Scegliamo un nome univoco per il file basato sul timestamp
		File dir = new File(Config.TMP_DIR);
		if(!dir.exists())
			dir.mkdirs();
		
		while(file == null || file.exists()){
			file = new File(Config.TMP_DIR + "/" + "AlloyInput" + 
					System.currentTimeMillis());
		}
		
		//scriviamo il contenuto
		FileWriter fw = new FileWriter(file, false);
		fw.write(contents);
		fw.close();
		
		return file;
	}

	/**
	 * 
	 * @param f
	 * @return
	 */
	private String getFileContents(File f, boolean printLineNumer){
		String contents = "";
		try{
			
			BufferedReader br = new BufferedReader(new InputStreamReader(
					new FileInputStream(f)));
			String line = null;
			line = br.readLine();
			for(int i = 1; line != null; i++){
				if(printLineNumer)
					contents += "\t" + i + ". ";
				else
					contents += "\t";
				
				contents += line + "\n"; 
				line = br.readLine();
			}
		}
		catch(Exception e){
			contents = "Cannot get file contents: " + e.toString();
		}
		return contents;
	}
	
	
	/**
	 * 
	 * @return
	 */
	private boolean isVariablePresent(String in, String var){
		String tokens[] = in.split("(\\s|\\.|-|:|<|>|=|\\&|\\[|\\]|\\||\\(|\\))");
		for(String s:tokens)
			if(s.equals(var))
				return true;
			
		return false;
	}
	
	/**
	 * 
	 * @param function
	 * @return
	 */
	private boolean isTypePresent(Function function, String type){
		if(function.getReturnType().equals(type))
			return true;
		for(String t: function.getParameterTypes())
			if(t.equals(type))
				return true;
				
		return false;
	}
	
	/**
	 * Genera l'header del file di input
	 * @param function
	 */
	private String generateHeader(Function function) {
		String ret = "";
		
		if(isTypePresent(function, "int"))
			ret += "open util/integer as " + Config.INTEGER_MODULE_IMPORT_NAME + "\n";
		if(isTypePresent(function, "int[]"))
			ret += "open util/sequniv as " + Config.SEQ_MODULE_IMPORT_NAME + "\n";
		
		return ret + "\n";
	}
	
	
	/**
	 * Genera le sigs a partire dalla segnatura della funzione da testare
	 * @param function
	 */
	private String generateInputSig(Function function) throws JamoteModelingException{
		
		//Prima di tutto generiamo la sig che rapppresenta gli argomenti
		//della funzione.
		String ret = "sig " + InputSig + "\n{ \n";
		String type = null; 
		Iterator<String> typesIterator = function.getParameterTypes().iterator();
		int i = 0;
		while(typesIterator.hasNext()){
			type = typesIterator.next();
			
			String quantifier = "";
			if(type.equals("int"))
				quantifier = "one";
			else if(type.equals("int[]"))
				quantifier = "lone";
			else
				throw new JamoteModelingException("Unexpected exception!!! Fix this bug!");
			
			ret += "\t " + ArgField + i + ": "+ quantifier +" " + this.translateType(type);
			i++;
			
			if(typesIterator.hasNext())
				ret += ",\n";
			else
				ret += "\n";
		}
		ret += "}\n\n";
		
		//Qui vanno inserite altre sigs se necessario...
		
		return ret;
	}
	
	/**
	 * @param function
	 */
	private String generateMetaSig(EquivalencePartition partition) throws JamoteModelingException{
		
		String ret = "one sig " + MetaSig + "\n{ \n";
		ret += "\t"+MaxIntField+": one Int,\n";
		ret += "\t"+MinIntField+": one Int,\n";
		ret += "\t"+ModelSizeField+": one Int\n";
		ret += "}\n";
		ret += "{\n";
		int maxInt = ((1 << (partition.getBitwidth() - 1)) - 1);
		ret += "\t"+MaxIntField+"="+maxInt+" && \n";
		ret += "\t"+MinIntField+"="+(-maxInt)+" && \n";
		int modelDimension = -1;
		if(partition.getModelDimension() <= maxInt)
			modelDimension = partition.getModelDimension();
			
		ret += "\t"+ModelSizeField+"="+modelDimension+"\n";
		ret += "}\n\n";
		
		return ret;
	}
	
	/**
	 * 
	 * @param preconditions
	 * @return
	 */
	private String generateSupportSigs(Function function) {
		String ret = "";
		
		if(isTypePresent(function, "int[]"))
		{
			ret += "sig JamoteArray{\n";
			ret += "\tarray: seq Int\n";
			ret += "}\n\n";
			
			ret += "fun at[a: JamoteArray, i:seq/Int]: Int{\n";
			ret += "\ta.array[i]\n";
			ret += "}\n\n";
			
			ret += "fun length[a: JamoteArray]: Int{\n";
			ret += "\t#a.array\n";
			ret += "}\n\n";
			
			ret += "pred hasIndex[a: JamoteArray, idx: seq/Int]{\n";
			ret += "\t	idx < #a.array && idx >= 0 \n";
			ret += "}\n\n";
		}
		
		return ret;
	}
	
	/**
	 * Genera i fact relativi alle precondizioni
	 * @param preconditions
	 * @return
	 */
	private String generatePreconditions(String preconditions) {
		String ret = "";
		if(! preconditions.trim().equals("")){
			ret = "" +
				"fact preconditions \n" +
				"{ \n" +
				"\t";
			if(isVariablePresent(preconditions, InputSig.toLowerCase()))
				ret += "\tall "+InputSig.toLowerCase()+":"+InputSig + "| ";
			if(isVariablePresent(preconditions, MetaSig.toLowerCase()))
				ret += "\tall "+MetaSig.toLowerCase()+":"+MetaSig + "| ";
			
			ret += "\t(" + preconditions + ")";
			ret += "\n} \n\n";
		}
		
		return ret;
	}
	
	/**
	 * Genera i pred corrispondenti alle specifiche di input delle classi di equivalenza
	 * @param partition
	 * @return
	 */
	private String generateInputSpecification(EquivalencePartition partition) {
		String ret = "";
		if(! partition.getInputSpecification().trim().equals("")){
			ret = "" +
				"pred inputSpecs \n" +
				"{ \n";
			if(isVariablePresent(partition.getInputSpecification(), InputSig.toLowerCase()))
				ret += "\tall "+InputSig.toLowerCase()+":"+InputSig + "|";
			
			if(isVariablePresent(partition.getInputSpecification(), MetaSig.toLowerCase()))
				ret += "\tall "+ MetaSig.toLowerCase()+":"+MetaSig + "| ";
			
			ret += "(" + partition.getInputSpecification() + ")";
			ret += "\n} \n\n";
		}
		else
			ret ="pred inputSpecs{}\n\n";
	
		return ret;
	}
	
	
	/**
	 *
	 * @param partition
	 * @return
	 */
	private String generateModelGenerationRun(EquivalencePartition partition) {
		String ret = "run inputSpecs for " + partition.getModelDimension() + " but " +
			partition.getBitwidth() +" int, exactly 1 " + InputSig+", exactly 1 " + MetaSig;
		
		return ret;
	}
	
	/**
	 * Trasforma un tipo generico in un tipo Alloy
	 * @return
	 */
	private String translateType(String t) throws JamoteModelingException
	{
		if(t.equals("int"))
			return "Int";
		
		if(t.equals("int[]"))
			return "JamoteArray";
		
		throw new JamoteModelingException("Not supported argument type: " + t);
	}

	/**
	 * Esegue Alloy analizer prendendo in input il file specificato
	 * 
	 * @param file
	 */
	private List<InputModel> executeAlloyToFindModels(File file,Function function,
			EquivalencePartition partition)	throws JamoteModelingException{
		try{
			List<InputModel> models = new ArrayList<InputModel>();
			JamoteReporter rep = new JamoteReporter(getReporter());
			Module world = CompUtil.parseEverything_fromFile(
					 rep, null, file.getPath());
		        
			//Selezioniamo un pò di opzioni...
			A4Options options = new A4Options();
			if(Config.ALLOY_SOLVER.equals("SAT4J"))
				options.solver = A4Options.SatSolver.SAT4J;
			else if(Config.ALLOY_SOLVER.equals("MINISAT"))
				options.solver = A4Options.SatSolver.MiniSatJNI;
			else if(Config.ALLOY_SOLVER.equals("ZCHAFF"))
				options.solver = A4Options.SatSolver.ZChaffJNI;
			else
				options.solver = A4Options.SatSolver.SAT4J;
			
			//Per ogni comando
			for (Command command: world.getAllCommands()) {
				//Eseguo il comando
				A4Solution ans = TranslateAlloyToKodkod.execute_command(rep, world.getAllReachableSigs(), command, options);
				
				//Processo le soluzioni se soddisfacibili
				for(int i = 0; i < partition.getMaxNumberOfTests();){
					if(! ans.satisfiable())
						break;
					
					InputModel m = concretizeModel(ans, function);
					if(!Config.ALLOW_DUPLICATE_MODELS){
						if(! models.contains(m)){
							models.add(m);
							i++;
						}
						else
							getReporter().log(Level.VERBOSE, "Duplicate model removed: " + m);
					}
					else{
						models.add(m);
						i++;
					}
					
					ans = ans.next();
				}
			}
			
			return models;
		}
		catch(Err e){
			getReporter().log(Level.ERROR, "Error in Alloy solving process", e);
			throw new JamoteModelingException("Error in Alloy solving process", e);
		}
		
	}
	
	/**
	 * Traduce la soluzione Alloy in una ModelInstance
	 * @param ans
	 * @return
	 */
	private InputModel concretizeModel(A4Solution ans, Function function) throws Err, JamoteModelingException{
		InputModel model = new InputModel();
		
		SafeList<Sig> sigs = ans.getAllReachableSigs();
		for(Sig s:sigs) {
			//Dobbiamo cercare la Sig "InputSig"
			if(s.label.equals("this/" + InputSig))
			{
				for(Field f:s.getFields()){
					//Di quale argomento si tratta?
					int argn = new Scanner(f.label).useDelimiter(ArgField).nextInt();
					
					AbstractObject argValue;
					if(function.getParameterTypes().get(argn).equals("int"))
						argValue = concretizeIntTuple(ans, f);
					else if(function.getParameterTypes().get(argn).equals("int[]"))
						argValue = concretizeIntArrayTuple(ans, f);
					else
						throw new JamoteModelingException("Unexpected exception!!! Fix this bug!");
					
					model.getObjects().put(new Integer(argn),argValue);
				}
			}
		}
		
		return model;
	}

	
	/**
	 * Trasforma una tupla rappresentante una sig Array in una ArrayList
	 * @param ans
	 * @param valueTuple
	 * @return
	 * @throws Err
	 */
	private AbstractObject concretizeIntArrayTuple(A4Solution ans, Field f) 
			throws JamoteModelingException,Err{
		//Facciamo un check sul tipo
		Type t = f.type;
		List<List<PrimSig>> psigs = t.fold();
		if(t.arity() != 2)
			throw new JamoteModelingException("Unexpected argument type in abstract model," +
					"arity is != 2, JamoteArray expected");
		
  		PrimSig ps1 = psigs.get(0).get(1);
		if(!ps1.label.equals("this/JamoteArray"))
			throw new JamoteModelingException("Unexpected argument type" +
					"["+ps1.label+ "]" +
					"in abstract model, JamoteArray expected");
		
		//Ora che il tipo è corretto estraiamo il valore dell'argomento 
		A4TupleSet tupleSet = (A4TupleSet) ans.eval(f);
		
		//l'argomento è null
		if(!tupleSet.iterator().hasNext())
			return new AbstractObject("int[]","int[]",null);
		//la prima sig è quella che rappresenta la Sig padre es. Output$0
		//la seconda deve essere un Array
		PrimSig arraySig = tupleSet.iterator().next().sig(1);
		String sigLabel = tupleSet.iterator().next().atom(1);
		
		Map<Integer, Integer> arrayMap = new HashMap<Integer, Integer>();
		Field arrayField = arraySig.getFields().iterator().next();
		tupleSet = (A4TupleSet) ans.eval(arrayField);
		for(A4Tuple valueTuple: tupleSet){
			if(valueTuple.atom(0).equals(sigLabel))
				arrayMap.put(new Integer(valueTuple.atom(1)), new Integer(valueTuple.atom(2)));
		}
		
		//Ora trasformiamo la Map in un array
		int ret[]  = new int[arrayMap.size()];
		for(int i = 0; i < arrayMap.size(); i++)
			ret[i] = arrayMap.get(new Integer(i)).intValue();
		
		return new AbstractObject("int[]","int[]",ret);
	}

	/**
	 * Trasforma una tupla rappresentate un Int in un Integer
	 * @return
	 */
	private AbstractObject concretizeIntTuple(A4Solution ans, Field f) 
			throws JamoteModelingException,Err
	{
		//Facciamo un check sul tipo
		Type t = f.type;
		List<List<PrimSig>> psigs = t.fold();
		if(t.arity() != 2)
			throw new JamoteModelingException("Unexpected argument type in abstract model," +
					"arity is != 2, single int expected");
		
		PrimSig ps = psigs.get(0).get(1);
		if(! ps.label.equals("Int"))
			throw new JamoteModelingException("Unexpected argument type["+ps.label+"]" +
					"in abstract model, Int expected");
		
		
		//Ora che il tipo è corretto estraiamo il valore dell'argomento 
		A4TupleSet tupleSet = (A4TupleSet) ans.eval(f);
		
		//Nota: nel nostro caso è sempre 1 valore se è un int
		if(tupleSet.iterator().hasNext())
		{
			A4Tuple valueTuple = tupleSet.iterator().next();
			//il primo atom è quello che rappresenta la Sig padre es. Output$0
			return new AbstractObject(valueTuple.atom(1),"int",Integer.parseInt(valueTuple.atom(1)));
		}
		else
			return new AbstractObject("int","int",null);
	}

	/**
	 * 
	 */
	public int checkOutput(Function function, String defines, EquivalencePartition partition,
			 InputModel inputModel, OutputModel ret) 
			throws JamoteModelingException{
		File file = null;
		try{
			if(!checkOutputModelIntLimits(ret, partition))
				return 2;
			
			//Creo il file di input da passare ad Alloy
			file = this.createAlloyCheckFile(function, partition, defines, inputModel,ret);
			
			getReporter().log(Level.DEBUG, "Generated Input file:\n--------\n"+
					getFileContents(file,false));
			
			//Configuro Alloy ed eseguo lo script appena creato
			if(executeAlloyToCheckOutput(file))
				return 0;
				
			return 1;
		}
		catch(JamoteModelingException je){
			getReporter().log(Level.ERROR, "An error occurred during checking of output...aborting", je);
			
			//Stampiamo il contenuto del file
			getReporter().log(Level.ERROR, "Printing Alloy check file: \n" + 
					getFileContents(file,true));
			
			throw je;
		}
		catch(IOException e){
			getReporter().log(Level.ERROR, "IO Error during checking process...aborting", e);
			throw new JamoteModelingException("IO Error during checking process", e);
		}
		finally{
			//Elimino il file temporaneo
			if(file != null)
				file.delete();
		}
	}
	
	/**
	 * Controlliamo se il risultato è entro i limiti del bitwidth di alloy
	 * @param ret
	 * @param partition
	 * @return
	 */
	private boolean checkOutputModelIntLimits(OutputModel ret,
			EquivalencePartition partition) throws JamoteModelingException {
		
		int maxInt = ((1 << (partition.getBitwidth() - 1)) - 1);
		
		if(ret.getType().equals("int")){
			Integer val = (Integer) ret.getValue();
			if(val.intValue() > maxInt || val.intValue() < -maxInt)
				return false;
		}
		else if(ret.getType().equals("int[]")){
			int[] val = (int[]) ret.getValue();
			if(val.length > maxInt || val.length < -maxInt)
				return false;
			
			for(int v: val)
				if(v > maxInt || v < -maxInt)
					return false;
		}
		else if(ret.wasThrown())
			return true;
		else
			throw new JamoteModelingException("Unexpected abstract model type!");
		
		return true;
	}


	/**
	 * 
	 * @param function
	 * @param outputSpecification
	 * @param inputModel
	 * @param obj
	 * @return
	 */
	private File createAlloyCheckFile(Function function,
			EquivalencePartition partition,String defines,InputModel inputModel, 
			 OutputModel obj) 
			throws IOException, JamoteModelingException{
		
		
		String contents = "";
		File file = null;
		
		
		getReporter().log(Level.VERBOSE, "Creating Alloy check file...");
		
		//genero l'header
		contents += this.generateHeader(function);
		
		contents += this.generateMetaSig(partition);
		
		//genero le sig di supporto
		contents += this.generateSupportSigs(function);
		
		//Genero le sigs a partire dalla segnatura della funzione da testare
		contents += this.generateInputSig(function);
		
		
		//Genero le sigs le eccezioni
		contents += this.generateExcSigs(function,obj);
		
		//aggiungo i defines
		contents += defines + "\n\n";
		
		//Genero le sigs per il return
		contents += this.generateRetSig(function);
		
		//Genero i fact corrispondenti ai valori di input dell'oggetto
		contents += this.generateInputValuesFact(inputModel);
		
		contents += this.generateOutputValue(obj);
		
		contents += this.generateOutputSpecs(partition);
		
		contents += this.generateCheckRun(partition, obj);
		
		//Scegliamo un nome univoco per il file basato sul timestamp
		File dir = new File(Config.TMP_DIR);
		if(!dir.exists())
			dir.mkdirs();
		while(file == null || file.exists()){
			file = new File(Config.TMP_DIR + "/" + "AlloyCheck" + 
					System.currentTimeMillis());
		}
		
		//scriviamo il contenuto
		FileWriter fw = new FileWriter(file, false);
		fw.write(contents);
		fw.close();
		
		return file;
	}

	/**
	 * 
	 * @param function
	 * @return
	 */
	private String generateExcSigs(Function function, OutputModel output){
		String ret = "sig " + ExcSig + "\n{}\n\n";
		
		for(String excName: function.getExceptionTypes())
			if(!excName.equals("Exception"))
				ret +=  "sig " + excName + " extends "+ ExcSig +"\n{}\n\n";
		
		if(output != null && output.wasThrown())
		{
			boolean presente = false;
			if(output.getType().equals("Exception"))
				presente = true;
			else
				for(String excName: function.getExceptionTypes())
					if(excName.equals(output.getType()))
						presente = true;
			
			if(!presente)
				ret +=  "sig " + output.getType() + " extends "+ ExcSig +"\n{}\n\n";
		}
		
		return ret;
	}
	
	
	/**
	 * 
	 * @param function
	 * @return
	 */
	private String generateRetSig(Function function) throws JamoteModelingException{
		String ret = "sig " + OutputSig + "\n{ \n" +
				RetField + ": lone " + translateType(function.getReturnType()) + ",\n"+
				ExcField + ": lone " + ExcSig + 
				"\n}\n\n";
		
		return ret;
	}

	/**
	 * 
	 * @param inputModel
	 * @return
	 */
	private String generateInputValuesFact(InputModel inputModel) throws JamoteModelingException{
		String ret = "";
		ret = "" +
			"fact inputValues \n" +
			"{ \n";
		for(int i = 0; i < inputModel.getObjects().size(); i++){
			AbstractObject o = inputModel.getObjects().get(new Integer(i));
			if(o.getType().equals("int")){
				ret += "(all a:" + InputSig + " | a." + ArgField +i+"="+o.getValue()+")\n";
				ret += "&& (all a:" + InputSig + " | one u:univ |a." + ArgField +i+"=u)\n";
			}
			else if(o.getType().equals("int[]")){
				ret += "(";
				if(o.getValue() == null)
				{
					ret += "(all a:" + InputSig + " | no u:univ | a." + ArgField +i+"=u)\n";
				}
				else
				{
					int[] arr = (int[]) o.getValue();
					ret += "all a:" + InputSig + " | a." + ArgField +i+".length="+arr.length;
					ret += " && ";
					for(int j = 0; j < arr.length; j++){
						ret += "a." + ArgField +i+".at["+j+"]="+arr[j] + " ";
						ret += " && ";
					}
					ret += "(all a:" + InputSig + " | one u:univ |a." + ArgField +i+"=u)\n";
				}
				
				ret += ")\n";
			}
			else 
				throw new JamoteModelingException("Unsupported abstract type "+ o.getType());
			
			if(inputModel.getObjects().containsKey(new Integer(i+1)))
				ret += " && ";
		}
		ret += "} \n\n";
		
		return ret;
	}

	/**
	 * 
	 * @param obj
	 * @return
	 */
	private String generateOutputValue(OutputModel o) throws JamoteModelingException{
		//Se è stato lanciato come un'eccezione
		if(o.wasThrown()){
			return "fact exceptionValue \n{\n" +
					"\t(all o:"+OutputSig+" | one e:"+ o.getType()+"| e=o."+ExcField+") && \n" +
					"\t(all o:"+OutputSig+" | no u:univ | o.return = u)\n" +
					"}\n\n";
		}
		
		String ret = "";
		ret = "" +
			"fact outputValues \n" +
			"{ \n";
		if(o.getType().equals("int")){
			ret += "(all r:" + OutputSig + " | r."+ RetField +"="+o.getValue()+")\n";
			ret += "&& (all r:Output | one u:univ |r.return = u)";
		}
		else if(o.getType().equals("int[]")){
			ret += "(";
			if(o.getValue() == null)
			{
				ret += "(all r:" + OutputSig + " | no u:univ | r." + RetField + "=u)\n";
			}
			else
			{
				
				int[] arr = (int[]) o.getValue();
				ret += "all r:" + OutputSig + " | r."+ RetField +".length = " + arr.length;
				if(arr.length > 0)
					ret += " && ";
				for(int j = 0; j < arr.length; j++){
					ret += "r."+ RetField +".at["+j+"]="+arr[j] + " ";
					ret += " && ";
				}
				
				ret += "(all r:Output | one u:univ |r.return = u)";
			}
			
			ret += ")\n";
		}
		else
			throw new JamoteModelingException("Unexpected error! Fix this bug!!!");
		
		ret += "&& (all e:"+ExcSig+" | no u:univ | e = u)\n";
		ret += "} \n\n";
		return ret;
	}
	
	/**
	 * 
	 * @param outputSpecification
	 * @return
	 */
	private String generateOutputSpecs(EquivalencePartition partition) {
		String ret = "assert outputSpecs\n{\n";
		if(isVariablePresent(partition.getOutputSpecification(), InputSig.toLowerCase()))
			ret += "\tall "+InputSig.toLowerCase()+":"+InputSig+ "|\n";
		if((isVariablePresent(partition.getOutputSpecification(), OutputSig.toLowerCase())))
			ret += "\tall "+OutputSig.toLowerCase()+":"+OutputSig+ "|\n";
		if((isVariablePresent(partition.getOutputSpecification(), MetaSig.toLowerCase())))
			ret += "\tall "+MetaSig.toLowerCase()+":"+MetaSig+ "|\n";
			
		ret += "(" + partition.getOutputSpecification() + ")" + "\n}\n\n";
		
		return ret;
	}
	
	private String generateCheckRun(EquivalencePartition partition, OutputModel out) {
		String ret = "check outputSpecs for " + partition.getModelDimension() + " but " + 
			partition.getBitwidth() + " int, exactly 1 " + InputSig + ",exactly 1 " + OutputSig +
			" ,1 " + ExcSig + ", exactly 1 " + MetaSig;
		
		return ret;
	}

	
	
	/**
	 * 
	 * @param file
	 * @return
	 * @throws JamoteModelingException
	 * @throws IOException
	 */
	private boolean executeAlloyToCheckOutput(File file) throws JamoteModelingException, IOException {
		try{
			JamoteReporter rep = new JamoteReporter(getReporter());
			Module world = CompUtil.parseEverything_fromFile(
					 rep, null, file.getPath());
		    
			//Selezioniamo un pò di opzioni...
			A4Options options = new A4Options();
			if(Config.ALLOY_SOLVER.equals("SAT4J"))
				options.solver = A4Options.SatSolver.SAT4J;
			else if(Config.ALLOY_SOLVER.equals("MINISAT"))
				options.solver = A4Options.SatSolver.MiniSatJNI;
			else if(Config.ALLOY_SOLVER.equals("ZCHAFF"))
				options.solver = A4Options.SatSolver.ZChaffJNI;
			else
				options.solver = A4Options.SatSolver.SAT4J;
			
			//Per ogni comando
			for (Command command: world.getAllCommands()) {
				//Eseguo il comando
				A4Solution ans = TranslateAlloyToKodkod.execute_command(rep, world.getAllReachableSigs(), command, options);
				return !ans.satisfiable();
			}
			
			throw new JamoteModelingException("Unexpected error, please fix this bug!");
		}
		catch(Err e){
			getReporter().log(Level.ERROR, "Error in Alloy solving process", e);
			throw new JamoteModelingException("Error in Alloy solving process", e);
		}
	}
	
	
	/**
	 * 
	 */
	public OutputModel getExpectedOutput(Function function, String defines, EquivalencePartition partition,
			InputModel inputModel) 
			throws JamoteModelingException{
		File file = null;
		
		try{
			//Creo il file di input da passare ad Alloy
			file = this.createAlloyExpOutFile(function, defines , partition, inputModel);
			
			getReporter().log(Level.DEBUG, "Generated Input file:\n--------\n"+
					getFileContents(file,false));
			
			//Configuro Alloy ed eseguo lo script appena creato
			return executeAlloyToGetExpectedOutput(file, function);
		}	
		catch(JamoteModelingException je){
			getReporter().log(Level.ERROR, "An error occurred during checking of output...aborting", je);
			
			//Stampiamo il contenuto del file
			getReporter().log(Level.ERROR, "Printing Alloy input file: \n" + 
					getFileContents(file,true));
			
			throw je;
		}
		catch(IOException e){
			getReporter().log(Level.ERROR, "IO Error during process...aborting", e);
			throw new JamoteModelingException("IO Error during process", e);
		}
		finally{
			//Elimino il file temporaneo
			if(file != null)
				file.delete();
		}
	}


	/**
	 * 
	 * @param function
	 * @param outputSpecification
	 * @param inputModel
	 * @param defines
	 * @param ret
	 * @param modelDimension
	 * @return
	 * @throws JamoteModelingException
	 * @throws IOException
	 */
	private File createAlloyExpOutFile(Function function,
			String defines, EquivalencePartition partition,  InputModel inputModel) 
			throws JamoteModelingException, IOException{
		
		String contents = "";
		File file = null;
		
		
		getReporter().log(Level.VERBOSE, "Creating Alloy expected output file...");
		
		//genero l'header
		contents += this.generateHeader(function);
		
		contents += this.generateMetaSig(partition);
		
		//genero le sig di supporto
		contents += this.generateSupportSigs(function);
		
		//Genero le sigs a partire dalla segnatura della funzione da testare
		contents += this.generateInputSig(function);
		
		contents += this.generateExcSigs(function, null);
		
		//aggiungo i defines
		contents += defines + "\n\n";
		
		//Genero le sigs per il return
		contents += this.generateRetSig(function);
		
		//Genero i fact corrispondenti ai valori di input dell'oggetto
		contents += this.generateInputValuesFact(inputModel);
		
		contents += this.generateExpectedOutputPred(partition);
		
		
		contents += this.generateExpectedOutputRun(partition);
		
		File dir = new File(Config.TMP_DIR);
		if(!dir.exists())
			dir.mkdirs();
		
		//Scegliamo un nome univoco per il file basato sul timestamp
		while(file == null || file.exists()){
			file = new File(Config.TMP_DIR + "/" + "AlloyExpOut" + 
					System.currentTimeMillis());
		}
		
		//scriviamo il contenuto
		FileWriter fw = new FileWriter(file, false);
		fw.write(contents);
		fw.close();
		
		return file;
	}

	/**
	 * 
	 * @param file
	 * @param function
	 * @return
	 * @throws JamoteModelingException
	 */
	private OutputModel executeAlloyToGetExpectedOutput(File file,Function function) throws JamoteModelingException {
		try{
			JamoteReporter rep = new JamoteReporter(getReporter());
			Module world = CompUtil.parseEverything_fromFile(
					 rep, null, file.getPath());
		        
			//Selezioniamo un pò di opzioni...
			A4Options options = new A4Options();
			if(Config.ALLOY_SOLVER.equals("SAT4J"))
				options.solver = A4Options.SatSolver.SAT4J;
			else if(Config.ALLOY_SOLVER.equals("MINISAT"))
				options.solver = A4Options.SatSolver.MiniSatJNI;
			else if(Config.ALLOY_SOLVER.equals("ZCHAFF"))
				options.solver = A4Options.SatSolver.ZChaffJNI;
			else
				options.solver = A4Options.SatSolver.SAT4J;
			
			//Per ogni comando
			for (Command command: world.getAllCommands()) {
				//Eseguo il comando
				A4Solution ans = TranslateAlloyToKodkod.execute_command(rep, world.getAllReachableSigs(), command, options);
				
				//Processo la soluzione se soddisfacibile
				if(ans.satisfiable())
					return processOutputModel(ans, function);
				
				return null;
			}
			
			return null;
		}
		catch(Err e){
			getReporter().log(Level.ERROR, "Error in Alloy solving process", e);
			throw new JamoteModelingException("Error in Alloy solving process", e);
		}
	}

	/**
	 * 
	 * @return
	 */
	private String generateExpectedOutputPred(EquivalencePartition partition) {
		String ret = "pred expectedOutput\n{\n";
		if(isVariablePresent(partition.getOutputSpecification(), InputSig.toLowerCase()))
			ret += "\tall "+InputSig.toLowerCase()+":"+InputSig+ "|\n";
		if((isVariablePresent(partition.getOutputSpecification(), OutputSig.toLowerCase())))
			ret += "\tall "+OutputSig.toLowerCase()+":"+OutputSig+ "|\n";
		if((isVariablePresent(partition.getOutputSpecification(), MetaSig.toLowerCase())))
			ret += "\tall "+MetaSig.toLowerCase()+":"+MetaSig+ "|\n";
			
		ret += "(" + partition.getOutputSpecification() + ")" + "\n}\n\n";
		
		return ret;
	}
	
	
	/**
	 * 
	 * @param partition
	 * @return
	 */
	private String generateExpectedOutputRun(EquivalencePartition partition) {
		String ret = "run expectedOutput for " + partition.getModelDimension() + " but " + 
		partition.getBitwidth() + " int, exactly 1 " + InputSig + ",exactly 1 " + OutputSig +
		", 1 " + ExcSig + ", exactly 1 " + MetaSig;
		return ret;
	}
	
	/**
	 * 
	 * @param ans
	 * @param function
	 * @return
	 * @throws Err
	 * @throws JamoteModelingException
	 */
	private OutputModel processOutputModel(A4Solution ans, Function function) 
			throws Err, JamoteModelingException
	{
		SafeList<Sig> sigs = ans.getAllReachableSigs();
		
		for(Sig s:sigs) {
			//Dobbiamo cercare la Sig "Output"
			if(s.label.equals("this/" + OutputSig))
			{
				//Prima cerchiamo l'eccezione, se c'è...
				for(Field f:s.getFields()){
					if(f.label.equals(ExcField)){
						OutputModel out = concretizeException(ans,f);
						if(out != null)
							return out;
					}
				}
				
				//Non ci sono eccezioni estraiamo gli argomenti
				for(Field f:s.getFields()){
					if(f.label.equals(RetField)){
						AbstractObject retValue;
						if(function.getReturnType().equals("int"))
							retValue = concretizeIntTuple(ans, f);
						else if(function.getReturnType().equals("int[]"))
							retValue = concretizeIntArrayTuple(ans, f);
						else
							throw new JamoteModelingException("Unexpected exception!!! Fix this bug!");
						
						return new OutputModel(retValue.getLabel(),retValue.getType(), retValue.getValue());
					}
				}
			}
		}
		
		return null;
	}

	/**
	 * 
	 * @param ans
	 * @param f
	 * @return
	 */
	private OutputModel concretizeException(A4Solution ans, Field f) throws 
			JamoteModelingException, Err
	{
		//Estraiamo il tipo
		Type t = f.type;
		List<List<PrimSig>> psigs = t.fold();
		if(t.arity() != 2)
			throw new JamoteModelingException("Unexpected argument type in abstract model," +
					"arity is != 2, Exception expected");
		
		//Ora che il tipo è corretto estraiamo il valore dell'argomento 
		A4TupleSet tupleSet = (A4TupleSet) ans.eval(f);
		
		//Se non è vuoto allora abbiamo un'eccezione
		if(tupleSet.iterator().hasNext())
			return new OutputModel(psigs.get(0).get(1).label,psigs.get(0).get(1).label,null,true);
		
		return null;
	}

}



/**
 * 
 */
class JamoteReporter extends A4Reporter{
	protected Reporter reporter;
	
	public JamoteReporter(Reporter reporter){
		super();
		this.reporter = reporter;
	}
	
    /**
     * This method is called at various points to report the current progress;
     * it is only intended as a debugging aid for the developers;
     * the messages are generally not useful for end users.
     *
     * @param msg - the debug message
     */
    //public void debug(String msg)
    //{}

	/** This method is called by the parser to report parser events. */
	//public void parse(String msg) { if (parent!=null) parent.parse(msg); }

    /** This method is called by the typechecker to report the type for each field/function/predicate/assertion, etc. */
    //public void typecheck(String msg) { if (parent!=null) parent.typecheck(msg); }

    /** This method is called by the typechecker to report a nonfatal type error. */
    public void warning(ErrorWarning msg){
    	reporter.log(Level.WARNING, "Alloy warning. Message: "+msg.msg + 
    			", Position: " + msg.pos.filename + "@" + msg.pos.y 
    			,msg.getCause());
    }

    /** This method is called by the ScopeComputer to report the scope chosen for each sig. */
    //public void scope(String msg) { if (parent!=null) parent.scope(msg); }

    /** This method is called by the BoundsComputer to report the bounds chosen for each sig and each field. */
    //public void bound(String msg) { if (parent!=null) parent.bound(msg); }

    /**
     * This method is called by the translator just before it begins generating CNF.
     *
     * @param solver - the solver chosen by the user (eg. SAT4J, MiniSat...)
     * @param bitwidth - the integer bitwidth chosen by the user
     * @param maxseq - the scope on seq/Int chosen by the user
     * @param skolemDepth - the skolem function depth chosen by the user (0, 1, 2...)
     * @param symmetry - the amount of symmetry breaking chosen by the user (0...)
     */
    public void translate(String solver, int bitwidth, int maxseq, int skolemDepth, int symmetry){
    	reporter.log(Level.VERBOSE, "Alloy is generating CNF. Solver: "+ solver +
    			", Integer bitwidth: " + bitwidth + "."
    			); 
    }

    /**
     * This method is called by the translator just after it generated the CNF.
     *
     * @param primaryVars - the total number of primary variables
     * @param totalVars - the total number of variables including the number of primary variables
     * @param clauses - the total number of clauses
     */
    public void solve(int primaryVars, int totalVars, int clauses){
    	reporter.log(Level.VERBOSE, "Alloy is searching for models. Total Vars: "+ totalVars +
    			", Clauses: " + clauses + "."
    			); 
    }

    /**
     * If solver==KK or solver==CNF, this method is called by the translator after it constructed the Kodkod or CNF file.
     *
     * @param filename - the Kodkod or CNF file generated by the translator
     */
    //public void resultCNF(String filename) { if (parent!=null) parent.resultCNF(filename); }

    /**
     * If solver!=KK && solver!=CNF, this method is called by the translator if the formula is satisfiable.
     *
     * @param command - this is the original Command used to generate this solution
     * @param solvingTime - this is the number of milliseconds the solver took to obtain this result
     * @param solution - the satisfying A4Solution object
     */
    public void resultSAT(Object command, long solvingTime, Object solution){
    	reporter.log(Level.VERBOSE, "The Alloy formula is satisfiable. Solving time: "+ solvingTime +
    			" ms."
    			); 
    }

    /**
     * If solver!=KK && solver!=CNF, this method is called by the translator before starting the unsat core minimization.
     *
     * @param command - this is the original Command used to generate this solution
     * @param before - the size of the core before calling minimization
     */
    //public void minimizing(Object command, int before) { if (parent!=null) parent.minimizing(command, before); }

    /**
     * If solver!=KK && solver!=CNF, this method is called by the translator after performing the unsat core minimization.
     *
     * @param command - this is the original Command used to generate this solution
     * @param before - the size of the core before calling minimization
     * @param after - the size of the core after calling minimization
     */
    //public void minimized(Object command, int before, int after) { if (parent!=null) parent.minimized(command, before, after); }

    /**
     * If solver!=KK && solver!=CNF, this method is called by the translator if the formula is unsatisfiable.
     *
     * @param command - this is the original Command used to generate this solution
     * @param solvingTime - this is the number of milliseconds the solver took to obtain this result
     * @param solution - the unsatisfying A4Solution object
     */
    public void resultUNSAT(Object command, long solvingTime, Object solution) {
    	reporter.log(Level.VERBOSE, "The Alloy formula is unsatisfiable. Solving time: "+ solvingTime +
    			" ms."
    			);
    }

    /** This method is called by the A4SolutionWriter when it is writing a particular sig, field, or skolem. */
    //public void write(Object expr) { if (parent!=null) parent.write(expr); }
}
